#!/bin/sh

test -f configure || { echo "change into the SDL-1.2 directory before running this script"; exit 1; }

# no dsymutil on Linux: normal configury would fail for _many_ things:
(cat configure > configure.cross && sed -i s/\"-g/\"/ configure.cross && chmod +x configure.cross) || exit 10

# Number of CPUs (for make -j)
NJOB=2

# dylib install name
INSTALL_NAME=@executable_path/../Frameworks/SDL.framework/Versions/A/SDL

# Export cross-toolchain paths
export PATH=/opt/cross_osx-ppc/bin:/opt/cross_osx-x86/bin:/opt/cross_osx-x86_64/usr/bin:$PATH
# use yasm for being capable of symbol visibility (i.e. 'function hidden')
export NASM=$HOME/bin/yasm

# Generic, cross-platform CFLAGS you always want go here.
CFLAGS="-O2 -Wall -pipe"

BASE_CONFIG_FLAGS="--disable-rpath --disable-static --disable-video-x11"

# PowerPC 32-bit compiler flags
CONFIG_PPC="--host=powerpc-apple-darwin9"
#  --x-includes=/opt/MacOSX10.4u.sdk/usr/X11R6/include --x-libraries=/opt/MacOSX10.4u.sdk/usr/X11R6/lib
BUILD_FLAGS_PPC="-mmacosx-version-min=10.4"

# Intel 32-bit compiler flags
CONFIG_X86="--host=i686-apple-darwin9"
#  --x-includes=/opt/MacOSX10.4u.sdk/usr/X11R6/include --x-libraries=/opt/MacOSX10.4u.sdk/usr/X11R6/lib
BUILD_FLAGS_X86="-mmacosx-version-min=10.4"

# Intel 64-bit compiler flags
CONFIG_X64="--host=x86_64-apple-darwin9"
#  --x-includes=/opt/cross_osx-x86_64/SDKs/MacOSX10.6.sdk/usr/X11R6/include --x-libraries=/opt/cross_osx-x86_64/SDKs/MacOSX10.6.sdk/usr/X11R6/lib
BUILD_FLAGS_X64="-mmacosx-version-min=10.6"
LFLAGS_X64="-Wl,-lbundle1.o"

# Create the build directories
rm -rf OSX
for dir in OSX OSX/ppc OSX/x86 OSX/x64; do
    mkdir $dir || exit 2
done

# Build the PowerPC 32-bit binary
echo "BUILD PPC:"
cd OSX/ppc
../../configure.cross $BASE_CONFIG_FLAGS $CONFIG_PPC CFLAGS="$CFLAGS $BUILD_FLAGS_PPC $CFLAGS_PPC" LDFLAGS="$BUILD_FLAGS_PPC $LFLAGS_PPC" || exit 10
(echo "#define HAVE_MEMCMP 1" >> include/SDL_config.h && echo "#define HAVE_STRTOD 1" >> include/SDL_config.h) || exit 11
make -j$NJOB || exit 12
cd ../..

# Build the Intel 32-bit binary
echo "BUILD X86:"
cd OSX/x86
../../configure.cross $BASE_CONFIG_FLAGS $CONFIG_X86 CFLAGS="$CFLAGS $BUILD_FLAGS_X86 $CFLAGS_X86" LDFLAGS="$BUILD_FLAGS_X86 $LFLAGS_X86" || exit 20
(echo "#define HAVE_MEMCMP 1" >> include/SDL_config.h && echo "#define HAVE_STRTOD 1" >> include/SDL_config.h) || exit 21
make -j$NJOB || exit 22
cd ../..

# Build the Intel 64-bit binary
echo "BUILD X64:"
cd OSX/x64
../../configure.cross $BASE_CONFIG_FLAGS $CONFIG_X64 CFLAGS="$CFLAGS $BUILD_FLAGS_X64 $CFLAGS_X64" LDFLAGS="$BUILD_FLAGS_X64 $LFLAGS_X64" || exit 30
(echo "#define HAVE_MEMCMP 1" >> include/SDL_config.h && echo "#define HAVE_STRTOD 1" >> include/SDL_config.h) || exit 31
make -j$NJOB || exit 32
cd ../..

cd OSX
# Strip any debug syms
powerpc-apple-darwin9-strip -S ppc/build/.libs/libSDL.dylib || exit 40
i686-apple-darwin9-strip -S x86/build/.libs/libSDL.dylib || exit 41
x86_64-apple-darwin9-strip -S x64/build/.libs/libSDL.dylib || exit 42
# Adjust install name
powerpc-apple-darwin9-install_name_tool -id $INSTALL_NAME ppc/build/.libs/libSDL.dylib || exit 50
i686-apple-darwin9-install_name_tool -id $INSTALL_NAME x86/build/.libs/libSDL.dylib || exit 51
x86_64-apple-darwin9-install_name_tool -id $INSTALL_NAME x64/build/.libs/libSDL.dylib || exit 52
# Combine into fat binary
x86_64-apple-darwin9-lipo -create -o libSDL.dylib x64/build/.libs/libSDL.dylib x86/build/.libs/libSDL.dylib ppc/build/.libs/libSDL.dylib || exit 55
cd ..

# DONE
